/*
 * ARM JTAG Fault Injector
 *
 * Author: Ingo Korb <ingo@akana.de>
 *
 * Copyright (C) 2011-2014 Department of Computer Science,
 * Design Automation of Embedded Systems Group,
 * Dortmund University of Technology, all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither the name of the copyright holder nor the names of its
 *      contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef TARGET_H
#define TARGET_H

#include <stdint.h>

typedef struct {
  uint32_t r[16];  // registers of current mode only!
  uint32_t cpsr;
  uint32_t spsr;
} arm_regs_t;

/**
 * target halt - halt target cpu
 * @regmask: bit mask of registers that should be read on halt
 *
 * This function halts the target CPU and saves some of its current-mode
 * registers. It does not check if the target CPU is actually running.
 * @regmask specifies the registers to be saved (1=r0, 2=r1, 4=r2...),
 * but r0, r15 and CPSR are always saved because they are clobbered
 * during halt/resume. Returns 0 if successful, nonzero if timed out.
 *
 * FIXME: CPSR usually doesn't need to be saved
 */
int target_halt(uint16_t regmask);

/**
 * target_resume - resumes target cpu
 *
 * This function resumes the target CPU. It does not check if the
 * target CPU is actually halted. @regmask specifies which registers
 * should be restored - the registers saved by target_halt will always
 * be restored, so anything but 0 there is only used to specifically
 * overwrite registers that weren't saved.
 */
void target_resume(void);

/* read/write words from the target memory space */
// FIXME: Add byte/halfword versions, larger reads/writes?
/* note: target_write needs R1, make sure it was saved by target_halt! */
uint32_t target_read_memory(uint32_t address);
void     target_write_memory(uint32_t address, uint32_t value);

/* read/update target registers - actually just copies from/to the internal copy */
void target_read_registers(arm_regs_t *regs);
void target_write_registers(arm_regs_t *regs);

/* immediately update the target registers */
// FIXME: Doesn't do CPSR
void target_write_registers_now(arm_regs_t *regs, uint16_t mask);

/* simulate FIQ */
/* will spit out an error without doing anything if the target is already in FIQ mode */
void target_fiq(void);

#endif
