/*
 * ARM JTAG Fault Injector
 *
 * Author: Andreas Heinig <andreas.heinig@gmx.de>
 *
 * Copyright (C) 2011-2014 Department of Computer Science,
 * Design Automation of Embedded Systems Group,
 * Dortmund University of Technology, all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither the name of the copyright holder nor the names of its
 *      contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include "inject.h"
#include "parameters.h"
#include "arm-opcode.h"
#include "injtypes.h"
#include "random.h"
#include "jtag.h"

extern volatile uint32_t manual_inject_addr;

extern inj_stat_t inj_stat;
extern uint32_t inj_flags;

extern unsigned int rate_mem_error;
extern unsigned int rate_reg_error;

extern unsigned int ival_mem_inject;
extern unsigned int ival_reg_inject;

static void report_error(arm_regs_t * regs, uint16_t type, uint16_t size, uint32_t location)
{
	//printf("\n\033[33mINJECT(%d): %08x, size: %d\033[0m\n", (int) type, (unsigned int)location, (int)size);
	uint32_t type_and_size = ((uint32_t)type << 16) | ((uint32_t)size);
	target_write_memory(ERROR_DEV_VBASE + 0, type_and_size);
	target_write_memory(ERROR_DEV_VBASE + 4, location);
	if(((regs->cpsr & ARM_MODE_MASK) == ARM_MODE_FIQ) || (regs->cpsr & 0x040))
		printf("\033[31mNOT REPORTING FAULT!\033[0m\n");
	else
		target_fiq();
}

static uint32_t mem_get_random_address(void)
{
	return (((uint32_t)random_get()) % RAM_INJ_SIZE) & 0xFFFFFFFC;
}

static int mem_caninject(uint32_t ptr)
{
	if(RAM_INJ_VBASE <= ptr && ptr < (RAM_INJ_VBASE + RAM_INJ_SIZE))
		return 1;
	else
		return 0;
}

static int mem_bitflip(arm_regs_t * regs, uint32_t addr, uint32_t size)
{
	/* reset counter and update statistic */
	ival_mem_inject -= rate_mem_error;

	inj_stat.mem_inject_cnt++;

	/* flip bit */
	if(inj_flags & ENABLE_FLIP)
	{
		uint32_t bit = ((uint32_t)random_get()) % (size * 8);
		uint32_t data = target_read_memory(addr + (bit / 32));
		data ^= 1 << (bit % 32);
		target_write_memory(addr + (bit / 32), data);
	}

	/* register fault */
	report_error(regs, ERROR_TYPE_MEMORY, size, addr);
	return 1;
}

static int reg_bitflip(arm_regs_t * regs, uint32_t reg)
{
	/* reset counter and update statistic */
	ival_reg_inject -= rate_reg_error;
	inj_stat.reg_inject_cnt++;

	/* flip bit */
	if(inj_flags & ENABLE_FLIP)
	{
		uint32_t bit = ((uint32_t)random_get()) % 32;
		regs->r[reg] ^= 1 << bit;
	}

	/* register fault */
	report_error(regs, ERROR_TYPE_REGISTER, 4, reg);
	return 1;
}


void inject(void)
{
	inj_stat.ival_inst_check_cnt++;
	if(!(inj_flags & ENABLE_INJECT))
		return;

	/* Stop target system */
	if(target_halt(0xFFFF))
	{
		printf("Will nicht!\n");
		return;
	}

	/* Read register file */
	arm_regs_t regs;
	target_read_registers(&regs);

	/* Get mode */
	int isModeFiq = (regs.cpsr & ARM_MODE_MASK) == ARM_MODE_FIQ ? 1 : 0;
	int isModeSys = (regs.cpsr & ARM_MODE_MASK) != ARM_MODE_USR ? 1 : 0;

	if((!isModeSys && (inj_flags & INJECT_ONLY_SYS_MODES)) ||
			((regs.cpsr & ARM_MODE_MASK) == ARM_MODE_UND)  ||
			((regs.cpsr & ARM_MODE_MASK) == ARM_MODE_ABT))
	{
		target_resume();
		jtag_idle();
		return;
	}

	/*
	 * only inject on specified address
	 */
	if(inj_flags & ADDRESS_ONLY)
	{
		if(isModeSys)
		{
			/* Don't inject unaware while in system mode */
			inj_stat.sys_cnt++;
			target_resume();
			jtag_idle();
			return;
		}
		mem_bitflip(&regs, manual_inject_addr & 0xFFFFFFFC, 4);
		target_resume();
		jtag_idle();
		return;
	}

	/*
	 * no instruction aware checking
	 */
	if(inj_flags & DISABLE_AWARENESS)
	{
		if(isModeSys)
		{
			/* Don't inject unaware while in system mode */
			inj_stat.sys_cnt++;
			target_resume();
			jtag_idle();
			return;
		}
		mem_bitflip(&regs, mem_get_random_address(), 4);
		target_resume();
		jtag_idle();
		return;
	}

	/***********************************
	 *                                 *
	 *   Standard injection routine    *
	 *                                 *
	 ***********************************/

	/*
	 * Get and decode executed instruction
	 */

	uint32_t inst = target_read_memory(regs.r[15]);
	arm_instruction_t inst_dec;
	arm_instruction_t * op = &inst_dec;

	if(verbose_opcode)
		printf("\nPC: %08x -- \033[31m", (unsigned int)regs.r[15]);
	int ret = decode_instruction(inst, &regs, op);
	if(verbose_opcode)
		printf("\033[0m");
	if(ret != 0)
	{
		if(ret == 2) /* instruction not executed */
		{
			target_resume();
			jtag_idle();
			return;
		}
		printf("Error while parsing instruction %08x @pc %08x\n", (unsigned int)inst, (unsigned int) regs.r[15]);
		target_resume();
		jtag_idle();
		return;
	}

	/*
	 * Filter registers
	 */

	/* PC is always reliable! */
	op->regs_r &= 0x7FFF;
	op->regs_w &= 0x7FFF;

	/* LR and SP are reliable in all system modes */
	if(isModeSys)
	{
		inj_stat.sys_cnt++;
		op->regs_r &= 0x9FFF;
		op->regs_w &= 0x9FFF;
	}

	/*
	 * Inject a register flip ?
	 */

	if((inj_flags & INJECT_TO_REG) &&
			(ival_reg_inject >= rate_reg_error) &&
			(op->regs_r != 0))
	{
		//inj_flags |= SINGLE_STEP_MODE | ENABLE_INJECT;
		int cnt, i;
		for(cnt = 0, i = 0; i < 16; i++)
			if(op->regs_r & (1 << i))
				cnt++;

		/* get register count */
		uint32_t reg = (cnt > 1 ? ((uint32_t)random_get()) % cnt : 0) + 1;
		for(i = 0; i < 16; i++)
		{
			if(op->regs_r & (1 << i))
				reg--;
			if(reg == 0)
			{
				reg_bitflip(op->regs, i);
				target_resume();
				jtag_idle();
				dump_instruction(op);
				return;
			}
		}
		printf("\nFATAL: I should not be here!: %s, %d\n", __func__, __LINE__);
		target_resume();
		jtag_idle();
		return;
	}

	/*
	 * Inject to memory ?
	 */
	if((inj_flags & INJECT_TO_MEM) &&
			(ival_mem_inject >= rate_mem_error) &&
			(op->flags & (OP_FLAG_READ | OP_FLAG_WRITE)))
	{
		if(op->mem_size < 4)
			op->mem_size = 4;

		if(op->flags & OP_FLAG_READ)
			inj_stat.inst_load_cnt++;
		if(op->flags & OP_FLAG_WRITE)
			inj_stat.inst_store_cnt++;

		uint32_t addr = op->mem_addr & 0xFFFFFFFC;

		/* are we allowed to inject at the address ? */
		if(!mem_caninject(addr))
		{
			target_resume();
			jtag_idle();
			return;
		}

		if(op->flags & OP_FLAG_READ)
			inj_stat.inst_load_noecc_cnt++;
		if(op->flags & OP_FLAG_WRITE)
		{
			inj_stat.inst_store_noecc_cnt++;
			/* only loads can cause errors */
			target_resume();
			jtag_idle();
			return;
		}

		mem_bitflip(op->regs, addr, op->mem_size);
		target_resume();
		jtag_idle();
		dump_instruction(op);
		return;
	}

	target_resume();
	jtag_idle();
}



