/*
 * ARM JTAG Fault Injector
 *
 * Author: Ingo Korb <ingo@akana.de>
 *
 * Copyright (C) 2011-2014 Department of Computer Science,
 * Design Automation of Embedded Systems Group,
 * Dortmund University of Technology, all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither the name of the copyright holder nor the names of its
 *      contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "hwconfig.h"
#include "timer.h"
#include "jtag.h"

#include "uart.h"

unsigned int jtag_clocks;
#define COUNT_CLOCKS(x) do {} while (0)
//#define COUNT_CLOCKS(x) jtag_clocks += (x)

/* --- macros/constants --- */

/* SSP status bits */
#define SSP_TFE 0   // Transmit FIFO empty
#define SSP_TNF 1   // Transmit FIFO not full
#define SSP_RNE 2   // Receive FIFO not empty
#define SSP_RFF 3   // Receive FIFO full
#define SSP_BSY 4   // Busy

#if SPI_ON_SSP == 0
#  define SSP_REGS LPC_SSP0
#  define SSP_PCLKREG PCLKSEL1
#  define SSP_PCLKBIT 10
#  define SSP_DMAID_TX 0
#  define SSP_DMAID_RX 1
#else
#  define SSP_REGS LPC_SSP1
#  define SSP_PCLKREG PCLKSEL0
#  define SSP_PCLKBIT 20
#  define SSP_DMAID_TX 2
#  define SSP_DMAID_RX 3
#endif

/* function in RAM for higher speed */
#ifdef FAST_FUNCTIONS_IN_RAM
#  define RAMTEXT __attribute__((section(".ramtext")))
#else
#  define RAMTEXT
#endif

static inline uint8_t reverse_byte(uint8_t in)
{
	uint8_t out;
	asm("rbit %[result], %[input]\n"
			"lsr  %[result], %[result], #24" : [result] "=r" (out) : [input] "r" (in));
	return out;
}

/* flip bit order in a word of up to 32 bit length */
static inline uint32_t flip_u32(uint32_t value, unsigned int num)
{
	uint32_t out;

	asm("rbit %[result], %[input]" : [result] "=r" (out) : [input] "r" (value));
	return out >> (32 - num);
}

/* --- internal state variables --- */

/* JTAG state names */
static char *statenames[] = {
		"UNKNOWN", "Test-Logic-Reset", "Run-Test/Idle", "Select-DR", "Capture-DR", "Shift-DR",
		"Exit1-DR", "Pause-DR", "Exit2-DR", "Update-DR",
		"Select-IR", "Capture-IR", "Shift-IR", "Exit1-IR", "Pause-IR", "Exit2-IR",
		"Update-IR"
};

/* currently selected scan chain */
static jtag_state_t current_state  = JTST_UNKNOWN;
static unsigned int current_chain  = -1;
static unsigned int current_ir     = -1;

// copied from OpenOCD
struct tms_sequence {
	uint8_t bits;
	uint8_t bit_count;
};

#define B8(bits,num) { 0b ## bits , (num) }

static const struct tms_sequence tms_sequences[6][6] = {
		/* to state: */
		/* RESET         IDLE           DRSHIFT        DRPAUSE        IRSHIFT        IRPAUSE */        /* from state: */
		{ B8(1111111,7), B8(0000000,7), B8(0010111,7), B8(0001010,7), B8(0011011,7), B8(0010110,7) },  /* RESET */
		{ B8(1111111,7), B8(0000000,7), B8(001,3),     B8(0101,4),    B8(0011,4),    B8(01011,5) },    /* IDLE */
		{ B8(1111111,7), B8(011,3),     B8(00111,5),   B8(01,2),      B8(001111,6),  B8(0101111,7) },  /* DRSHIFT */
		{ B8(1111111,7), B8(011,3),     B8(01,2),      B8(0,1),       B8(001111,6),  B8(0101111,7) },  /* DRPAUSE */
		{ B8(1111111,7), B8(011,3),     B8(00111,5),   B8(010111,6),  B8(001111,6),  B8(01,2) },       /* IRSHIFT */
		{ B8(1111111,7), B8(011,3),     B8(00111,5),   B8(010111,6),  B8(01,2),      B8(0,1)}          /* IRPAUSE */
};

/* --- SSP --- */

static inline void ssp_set_wordlen(unsigned int len)
{
	SSP_REGS->CR0 = (SSP_REGS->CR0 & ~0b111) | (len-1);
}

/* connect SSP to pins */
static inline void ssp_connect(void)
{
	ssp_set_wordlen(8);
#if SPI_ON_SSP == 0
	LPC_PINCON->PINSEL0 |= BV(31);        // SCK
	LPC_PINCON->PINSEL1 |= BV(3) | BV(5); // MISO/MOSI
#else
	LPC_PINCON->PINSEL0 |= BV(15) | BV(17) | BV(19);
#endif
}

/* disconnect SSP from pins */
static inline void ssp_disconnect(void)
{
#if SPI_ON_SSP == 0
	LPC_PINCON->PINSEL0 &= ~ BV(31);         // SCK
	LPC_PINCON->PINSEL1 &= ~(BV(3) | BV(5)); // MISO/MOSI
#else
	LPC_PINCON->PINSEL0 &= ~(BV(15) | BV(17) | BV(19));
#endif
}

/* --- various helper functions --- */

// FIXME: not side-effect-safe
#define min(x,y) ((x)<(y)?(x):(y))

// adapted from OpenOCD
static RAMTEXT int tap_move_ndx(unsigned int astate)
{
	/* given a stable state, return the index into the tms_seqs[]
	 * array within tap_get_tms_path()
	 */

	int ndx;

	switch (astate) {
	case JTST_RESET:     ndx = 0;            break;
	case JTST_IDLE:      ndx = 1;            break;
	case JTST_SHIFT_DR:  ndx = 2;            break;
	case JTST_PAUSE_DR:  ndx = 3;            break;
	case JTST_SHIFT_IR:  ndx = 4;            break;
	case JTST_PAUSE_IR:  ndx = 5;            break;
	default:
		printf("Illegal state in tap_move_ndx: %d\n", astate);
		while (1) ;
	}

	return ndx;
}

static inline void jtag_delay(void)
{
	JTAG_DELAY();
}

/* just toggle clock */
static RAMTEXT void jtag_clock_fast(void)
{
	COUNT_CLOCKS(1);
	jtag_delay();
	set_tck(1);
	jtag_delay();
	set_tck(0);
}

/* toggle clock, read TDI when it's high */
static RAMTEXT unsigned int jtag_clock_capture(void)
{
	unsigned int res;

	COUNT_CLOCKS(1);
	jtag_delay();
	set_tck(1);
	res = get_tdi();

	jtag_delay();
	set_tck(0);

	return res;
}

/* sets TMS and toggles clock once */
static RAMTEXT void jtag_clock(int tms)
{
	set_tms(tms);
	jtag_clock_fast();
}

/* shifts <bits> bits, sets ups everything for the last bit but doesn't clock it */
static RAMTEXT uint32_t jtag_shift_bits(uint32_t value, unsigned int bits)
{
	uint32_t result = 0;
	unsigned int i;

	for (i=bits; i!=1; i--) {
		// LSB first, confirmed in ARM docs
		set_tdo(value & 1);
		value >>= 1;
		jtag_delay();

		result >>= 1;
		result |= (get_tdi() << (bits - 1));

		COUNT_CLOCKS(1);
		set_tck(1);
		jtag_delay();
		set_tck(0);
	}

	set_tdo(value & 1);
	result >>= 1;
	jtag_delay();
	// technically this is too early because the value should be read at the rising edge and not before it
	result |= get_tdi();
	return result;
}


/* --- exported functions --- */

/* run through JTAG state machine to reach target */
/* if force is false, do nothing if already at target */
/* if force is true, run through Idle if already at target */
RAMTEXT void jtag_change_state(jtag_state_t target, unsigned int force)
{
	if (current_state == target)
	{
		if (force) {
			jtag_change_state(JTST_IDLE, 0);
			jtag_change_state(target, 0);
		}
		return;
	}

	uint8_t path, bitlen;
	unsigned int i;

	path   = tms_sequences[tap_move_ndx(current_state)][tap_move_ndx(target)].bits;
	bitlen = tms_sequences[tap_move_ndx(current_state)][tap_move_ndx(target)].bit_count;

	for (i=0; i<bitlen; i++)
	{
		jtag_clock(path & (1 << i));
	}

	current_state = target;
}

void jtag_init(void)
{
	/* configure TRST, SRST, TMS, TCK, TDO as output, high except TCK */
	LPC_GPIO0->FIOSET  =
			BV(JTAG_BIT_TRST) |
			BV(JTAG_BIT_TMS)  |
			BV(JTAG_BIT_SRST) |
			BV(JTAG_BIT_TDO);
	LPC_GPIO0->FIOCLR  = BV(JTAG_BIT_TCK);
	LPC_GPIO0->FIODIR |=
			BV(JTAG_BIT_TRST) |
			BV(JTAG_BIT_TMS)  |
			BV(JTAG_BIT_SRST) |
			BV(JTAG_BIT_TCK)  |
			BV(JTAG_BIT_TDO);

	/* SRST is open-drain */
	LPC_PINCON->PINMODE_OD0 |= BV(JTAG_BIT_SRST);

	/* configure SSP */
	/* Set clock prescaler to 1:1 */
	BITBAND(LPC_SC->SSP_PCLKREG, SSP_PCLKBIT) = 1;

	/* configure data format - 8 bits, SPI, CPOL=0, CPHA=0, 1 clock per bit */
	SSP_REGS->CR0 = (8-1);

	/* set clock prescaler */
	SSP_REGS->CPSR = SSP_CLK_DIVISOR;

	/* Enable SSP */
	SSP_REGS->CR1 = BV(1);
}

void jtag_idle(void)
{
	unsigned int i;

	/* invalidate "cached" values */
	current_ir = -1;
	current_chain = -1;

	/* send 20 clocks with TMS=1 to reset the state machine */
	set_tms(1);
	for (i=0; i<20; i++)
	{
		jtag_delay();
		set_tck(1);
		jtag_delay();
		set_tck(0);
	}

	/* move to Run-Test/Idle */
	jtag_clock(0);

	current_ir = -1;
	current_chain = -1;
	current_state = JTST_IDLE;
}

RAMTEXT uint32_t jtag_scan_ir(uint32_t value, unsigned int bits, jtag_state_t endstate) {
	uint32_t result = 0;
	unsigned int i;

	current_ir = value;

	/* move to Shift-IR */
	jtag_change_state(JTST_SHIFT_IR, 0);

	for (i=bits; i>1; i--)
	{
		// LSB first, confirmed in ARM docs
		set_tdo(value & 1);
		value >>= 1;

		result >>= 1;
		result |= (jtag_clock_capture() << (bits - 1));
	}

	/* final bit */
	set_tdo(value & 1);
	jtag_delay();
	result >>= 1;
	result |= (get_tdi() << (bits - 1));

	jtag_change_state(endstate, 0);

	return result;
}

/* set IR, skip if the instruction is already set */
RAMTEXT void jtag_set_ir(uint32_t value, unsigned int bits, jtag_state_t endstate)
{
	if (current_ir == value)
		return;

	jtag_scan_ir(value, bits, endstate);
}

RAMTEXT uint32_t jtag_scan_dr_single(uint32_t value, unsigned int bits, jtag_state_t endstate)
{
	uint32_t result = 0;
	unsigned int i;

	jtag_change_state(JTST_SHIFT_DR, 0);

	for (i=bits; i>1; i--)
	{
		set_tdo(value & 1);
		value >>= 1;

		result >>= 1;
		result |= (jtag_clock_capture() << (bits - 1));
	}

	/* final bit */
	set_tdo(value & 1);
	jtag_delay();
	result >>= 1;
	result |= (get_tdi() << (bits - 1));

	jtag_change_state(endstate, 0);

	return result;
} 

RAMTEXT void jtag_select_chain(unsigned int ch, jtag_state_t endstate)
{
	if (current_chain == ch)
		return;

	jtag_scan_ir(JTAG_SCAN_N, JTAG_IR_LEN, endstate);
	jtag_scan_dr_single(ch, 5, endstate);
	current_chain = ch;
}

RAMTEXT void jtag_scan_dr(unsigned int count, jtag_block_t *blocks, jtag_state_t endstate)
{
	jtag_block_t *curbl = blocks;
	unsigned int bits_rem, bits = 0, i;
	uint8_t vin = 0, vout = 0;
	uint8_t *in = NULL, *out;

	jtag_change_state(JTST_SHIFT_DR, 0);

	while (count--)
	{
		bits_rem = curbl->bits;
		in       = curbl->input;
		out      = curbl->output;

		/* hold back the final bit */
		if (count == 0)
			bits_rem -= 1;

		/* send as much as possible with hardware SPI */
		if (bits_rem >= 4)
		{
			ssp_connect();

			if (bits_rem >= 8)
			{
				// FIXME: Assumes that at most 64 bits are sent
				/* push data into transmit FIFO */
				unsigned int bytes = 0;
				while (bits_rem >= 8)
				{
					if (out)
						vout = reverse_byte(*out++); //bit_reverse_table256[*out++];
					else
						vout = 0;

					SSP_REGS->DR = vout;

					bytes++;
					bits_rem -= 8;
					COUNT_CLOCKS(8);
				}

				/* read data from RX FIFO */
				while (bytes > 0) {
					/* wait until data is available */
					while (!BITBAND(SSP_REGS->SR, SSP_RNE)) ;
					vin = SSP_REGS->DR;

					if (in)
						*in++ = reverse_byte(vin); //bit_reverse_table256[vin];

					bytes--;
				}
			}

			if (bits_rem >= 4)
			{
				/* finish the partial byte with hardware if possible */

				/* read last output byte */
				if (out)
					vout = *out++;
				else
					vout = 0;
				vin = 0;

				/* send partial bytes if possible */
				ssp_set_wordlen(bits_rem);

				SSP_REGS->DR = flip_u32(vout, bits_rem);
				while (!BITBAND(SSP_REGS->SR, SSP_RNE)) ;
				vin = flip_u32(SSP_REGS->DR, bits_rem);

				vout >>= bits_rem; // move final bit into position

				if (in)
					*in++ = vin;

				bits_rem = 0;
			}

			ssp_disconnect();
		}

		/* send the rest with bit-banging */
		if (bits_rem > 0)
		{
			// FIXME: assert(bits_rem <= 8)

			/* read last output byte */
			if (out)
				vout = *out++;
			else
				vout = 0;
			vin = 0;

			/* send the last bits */
			for (i=0; i<bits_rem; i++)
			{
				set_tdo(vout & 1);
				vout >>= 1;

				vin |= jtag_clock_capture() << i;
			}

			/* store result */
			if (in)
				*in++ = vin;
		}

		/* next block */
		curbl++;
	}

	/* final bit */
	// FIXME: I think this will fail for transfers with (len % 8) == 1 ?
	set_tdo(vout & 1);
	jtag_delay();
	vin |= get_tdi() << bits;
	if (in)
		*(in-1) = vin;

	jtag_change_state(endstate, 0);
}
