/*
 * ARM JTAG Fault Injector
 *
 * Author: Andreas Heinig <andreas.heinig@gmx.de>
 *
 * Copyright (C) 2011-2014 Department of Computer Science,
 * Design Automation of Embedded Systems Group,
 * Dortmund University of Technology, all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither the name of the copyright holder nor the names of its
 *      contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <arm/NXP/LPC17xx/LPC17xx.h>
#include <arm/bits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "hwconfig.h"
#include "jtag.h"
#include "target.h"
#include "system.h"
#include "timer.h"
#include "uart.h"
#include "touchscreen.h"
#include "random.h"
#include "inject.h"

#include "embice.h"
#include "injtypes.h"
#include "parameters.h"
#include "arm-opcode.h"

void print_stat(void);
void print_stat_header(void);

volatile uint32_t manual_inject_addr = RAM_INJ_VBASE;

inj_stat_t inj_stat;
uint32_t inj_flags = INJECT_DEFAULT_FLAGS;

volatile unsigned int rate_inst_check = DEFAULT_PARAM_INST_CHECK_RATE;
volatile unsigned int rate_mem_error  = DEFAULT_PARAM_MEM_ERROR_RATE;
volatile unsigned int rate_reg_error  = DEFAULT_PARAM_REG_ERROR_RATE;

volatile unsigned int ival_inst_check = 0;
volatile unsigned int ival_mem_inject = 0;
volatile unsigned int ival_reg_inject = 0;

static void clear_screen(void)
{
	printf("\e[2J");
}

static void print_key_codes(void)
{
	printf("\n==========  JTAG inject  ==========\n\n");
	printf("Key codes:\n");
	printf("   H ... Print this help message\n");
	printf("   C ... Clear statistics\n");
	printf("   V ... Enable verbosity\n");
	printf("\n");
	printf("   S ... Start injection\n");
	printf("   Q ... Stop injection\n");
	printf("   T ... Manual injection\n");
	printf("   O ... Inject into sysmodes only\n");
	printf("\n");
	printf("   M ... Enable memory fault injection\n");
	printf("   R ... Enable register fault injection\n");
	printf("   F ... Arm flip of bits (in memory and registers)\n");
	printf("   B ... Record access to memory blocks\n");
	printf("\n");
	printf("   U ... Use mem map\n");
	printf("   W ... Flip only on address A\n");
	printf("   X ... Disable instruction aware flipping\n");
	printf("\n");
	printf("   A ... Enter address\n");
	printf("   I ... Enter instruction check interval\n");
	printf("   J ... Enter memory flip interval\n");
	printf("   K ... Enter register flip interval\n");
	printf("\n");
	printf("   Y,Z ... Undocumented Options\n");
	printf("\n\n");

	printf("Tick:                       %4d ms\n", PARAM_TICK_INTERVAL);
	printf("Instruction check interval: %4d ticks (%d ms)\n", rate_inst_check, rate_inst_check * PARAM_TICK_INTERVAL);
	printf("Memory flip interval:       %4d ticks (%d ms)\n", rate_mem_error,  rate_mem_error  * PARAM_TICK_INTERVAL);
	printf("Register flip interval:     %4d ticks (%d ms)\n", rate_reg_error,  rate_reg_error  * PARAM_TICK_INTERVAL);
	printf("Address:              0x%08x\n", (unsigned int)manual_inject_addr);

	printf("\n\n");
	print_stat_header();
}

static void clear_all(void)
{
	memset(&inj_stat, 0, sizeof(inj_stat_t));
}

static void data_init(void)
{
	clear_all();
	random_init();
}

static void dump_registers(void)
{
	/* Stop target system */
	if(target_halt(0xFFFF))
	{
		printf("Failed to halt target!\n");
		return;
	}
	inj_stat.ival_inst_check_cnt++;

	/* Read register file */
	arm_regs_t regs;
	target_read_registers(&regs);
	target_resume();

	uint32_t * regptr = (uint32_t *)&regs;
	int i;
	printf("\n");
	for(i = 0; i < 16; i++)
		printf("R%-2d:  0x%08x (%d)\n", i, (unsigned int)regptr[i], (int)regptr[i]);
	printf("CPSR: 0x%08x (%d)\n", (unsigned int)regptr[i], (int)regptr[i]);
}

static void skip_instruction(void)
{
	/* Stop target system */
	if(target_halt(0xFFFF))
	{
		printf("Failed to halt target!\n");
		return;
	}

	/* Read register file */
	arm_regs_t regs;
	
	target_read_registers(&regs);
	regs.r[15] += 4;
	target_write_registers(&regs);
	target_resume();
}

static void read_debug_registers(void)
{
	/* Stop target system */
	if(target_halt(0xFFFF))
	{
		printf("Failed to halt target!\n");
		return;
	}

	/* Read register file */
	uint32_t data = target_read_memory(ERROR_DEV_VBASE + 8);
	printf("\nDBG:%08x (%d)\n", (unsigned int)data, (int)data);
	target_resume();
}

int main(void)
{
	/* Initialize */
	system_init_early();
	uart_init();
	timer_init();
	jtag_init();
	system_init_late();
	enable_interrupts();
	data_init();
	touchscreen_init();

	clear_screen();

	/* check the idcode of the target until it matches the expected one three times in a row */
	unsigned int counter = 0;
	while (counter < 3)
	{
		uint32_t idcode;
		jtag_idle();
		idcode = jtag_scan_dr_single(0, 32, JTST_IDLE);
		printf("idcode: %08lx\n", idcode);
		printf("initial IR: %08lx\n",
				jtag_scan_ir(0xf, 4, JTST_IDLE));
		printf("bypassed DR (shifted): %08lx\n",
				jtag_scan_dr_single(0x12345678, 32, JTST_IDLE) >> 1);

		if (idcode != 0x20a023d3)
		{
			printf("wrong idcode, retrying...\n");
			counter = 0;
		}
		else
		{
			counter++;
		}
		delay_ms(100);
	}

	clear_screen();

	print_key_codes();

	/*
	 * MAIN LOOP
	 */
	while(1)
	{
		delay_ms(PARAM_TICK_INTERVAL * PARAM_MAIN_DELAY);
		random_fill();

		/* check for events */
		if(uart_gotc())
		{
			unsigned char c = uart_getc();
			switch(c)
			{
			case 'h':
			case 'H':
				print_key_codes();
				break;
			case 'c':
			case 'C':
				clear_all();
				break;
			case 'm':
			case 'M':
				inj_flags ^= INJECT_TO_MEM;
				inj_flags &= ~RECORD_MEM;
				break;
			case 'b':
			case 'B':
				inj_flags ^= RECORD_MEM;
				inj_flags &= ~INJECT_TO_MEM;
				break;
			case 'r':
			case 'R':
				inj_flags ^= INJECT_TO_REG;
				break;
			case 'f':
			case 'F':
				inj_flags ^= ENABLE_FLIP;
				break;
			case 'o':
			case 'O':
				inj_flags ^= INJECT_ONLY_SYS_MODES;
				break;
			case 's':
			case 'S':
				ival_mem_inject = 0;
				ival_reg_inject = 0;
				inj_flags |= ENABLE_INJECT;
				inj_flags &= ~(SINGLE_STEP_MODE);
				break;
			case 'q':
			case 'Q':
				inj_flags &= ~(ENABLE_INJECT | SINGLE_STEP_MODE);
				break;
			case 't':
			case 'T':
				if(inj_flags & SINGLE_STEP_MODE)
					inject();
				else
					inj_flags |= SINGLE_STEP_MODE | ENABLE_INJECT;
				break;
			case 'v':
			case 'V':
				if(verbose_opcode)
					verbose_opcode = 0;
				else
					verbose_opcode = 1;
				break;
			case 'w':
			case 'W':
				inj_flags ^= ADDRESS_ONLY;
				inj_flags &= ~(DISABLE_AWARENESS);
				break;
			case 'x':
			case 'X':
				inj_flags ^= DISABLE_AWARENESS;
				inj_flags &= ~(ADDRESS_ONLY);
				break;
			case 'a':
			case 'A':
			{
				char buffer[32];
				char * endptr;
				printf("\nEnter address: ");
				uart_get_string(32, buffer);
				manual_inject_addr = (uint32_t)strtoul(buffer, &endptr, 0);
				printf("\n OK, using 0x%08x\n\n", (unsigned int)manual_inject_addr);
				break;
			}

			case 'i':
			case 'I':
			{
				char buffer[32];
				char * endptr;
				printf("\nEnter instruction check interval [ticks]: ");
				uart_get_string(32, buffer);
				rate_inst_check = (uint32_t)strtoul(buffer, &endptr, 0);
				printf("\n OK, using %d\n\n", rate_inst_check);
				break;
			}
			case 'j':
			case 'J':
			{
				char buffer[32];
				char * endptr;
				printf("\nEnter memory flip interval [ticks]: ");
				uart_get_string(32, buffer);
				rate_mem_error = (uint32_t)strtoul(buffer, &endptr, 0);
				printf("\n OK, using %d\n\n", rate_mem_error);
				break;
			}
			case 'k':
			case 'K':
			{
				char buffer[32];
				char * endptr;
				printf("\nEnter register flip interval [ticks]: ");
				uart_get_string(32, buffer);
				rate_reg_error = (uint32_t)strtoul(buffer, &endptr, 0);
				printf("\n OK, using %d\n\n", rate_reg_error);
				break;
			}

			case 'y':
			case 'Y':
				//inj_flags ^= OPTION_Y;
				//skip_instruction();
				read_debug_registers();
				break;
			case 'z':
			case 'Z':
				//inj_flags ^= OPTION_Z;
				dump_registers();
				break;
			default:
				break;
			}
			print_stat();

		}

		/* print statistics */
		static int stat_cnt = 0;
		stat_cnt++;
		if(stat_cnt > 10)
		{
			print_stat();
			stat_cnt = 0;
		}
	}
	return 0;
}

void print_stat(void)
{
	/* print flags */
	printf("\r");
	if(inj_flags & INJECT_TO_REG)
	{
		if(ival_reg_inject >= rate_reg_error)
			printf("\033[31mR\033[0m");
		else
			printf("R");
	}
	else
		printf(" ");
	if(inj_flags & INJECT_TO_MEM)
	{
		if(ival_mem_inject >= rate_mem_error)
			printf("\033[31mM\033[0m");
		else
			printf("M");
	}
	else if(inj_flags & RECORD_MEM)
		printf("A");
	else
		printf(" ");
	if(inj_flags & ENABLE_FLIP)
		printf("F");
	else
		printf(" ");
	if(inj_flags & SINGLE_STEP_MODE)
		printf("T");
	else if(inj_flags & ENABLE_INJECT)
		printf("S");
	else
		printf("Q");

	if(inj_flags & INJECT_ONLY_SYS_MODES)
		printf("O");
	else
		printf(" ");

	if(inj_flags & DISABLE_AWARENESS)
		printf("X");
	else if(inj_flags & ADDRESS_ONLY)
		printf("W");
	else
		printf(" ");

	if(verbose_opcode)
		printf("V");
	else
		printf(" ");

	if(inj_flags & OPTION_Y)
		printf("Y");
	if(inj_flags & OPTION_Z)
		printf("Z");

	/* stats */
	printf(" %10u %10u %10u %10u %10u %10u %10u %6u",
			(unsigned int)inj_stat.ival_inst_check_cnt,
			(unsigned int)inj_stat.inst_load_cnt,
			(unsigned int)inj_stat.inst_load_noecc_cnt,
			(unsigned int)inj_stat.inst_store_cnt,
			(unsigned int)inj_stat.inst_store_noecc_cnt,
			(unsigned int)inj_stat.mem_inject_cnt,
			(unsigned int)inj_stat.reg_inject_cnt,
			(unsigned int)inj_stat.sys_cnt);
}

void print_stat_header(void)
{
 	//              1234567890 1234567890 1234567890 1234567890 1234567890 1234567890 1234567890
	printf("FLAG         #HALT   #LD(ALL) #LD(NOECC)   #ST(ALL) #ST(NOECC) #M-INJECTS #R-INJECTS   #Sys\n");
}

/* tick hook => called every PARAM_TICK_INTERVAL ms */
void __attribute__((noinline)) SysTick_Hook(void)
{
	/* increment interval counters */
	ival_inst_check++;
	ival_mem_inject++;
	ival_reg_inject++;

	/* instruction check interval */
	if((inj_flags & ADDRESS_ONLY) || (inj_flags & DISABLE_AWARENESS))
	{
		if((ival_mem_inject >= rate_mem_error) && !(inj_flags & SINGLE_STEP_MODE))
		{
			ival_mem_inject = 0;
			inject();
		}
	}
	else
	{
		if(ival_inst_check >= rate_inst_check && !(inj_flags & SINGLE_STEP_MODE))
		{
			ival_inst_check = 0;
			inject();
		}
	}
	if (coord_updated)
	{
		uint32_t coord = touchscreen_get();
		embice_write_register(EMBICE_DEBUG_COMM_DATA, coord);
		touchscreen_toggle_int();
	}
}
