/*
 * ARM JTAG Fault Injector
 * - startup code for LPC17xx
 *
 * Author: Ingo Korb <ingo@akana.de>
 *
 * Copyright (C) 2011-2014 Department of Computer Science,
 * Design Automation of Embedded Systems Group,
 * Dortmund University of Technology, all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither the name of the copyright holder nor the names of its
 *      contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "hwconfig.h"
#include <arm/NXP/LPC17xx/LPC17xx.h>
#include <arm/bits.h>
#include "system.h"

/* Early system initialisation */
void system_init_early(void)
{
	__disable_irq();

	/* mbed LEDs */
	LPC_GPIO1->FIODIR |= BV(18) | BV(20) | BV(21) | BV(23);
	LPC_GPIO1->FIOPIN &= ~(BV(18) | BV(20) | BV(21) | BV(23));

	/* UART: use internal debug port on mbed */
	LPC_PINCON->PINSEL0 |= BV(4) | BV(6);
}

/* Late initialisation, increase CPU clock */
void system_init_late(void)
{
	/* Set flash accelerator to 5 CPU cycle access time */
	LPC_SC->FLASHCFG = (LPC_SC->FLASHCFG & 0xffff0fff) | (4 << 12);

	/* Enable main oscillator, range 1-20MHz */
	BITBAND(LPC_SC->SCS, 5) = 1;

	/* Wait until stable */
	while (!BITBAND(LPC_SC->SCS, 6)) ;

	/* Use main oscillator as system clock source */
	LPC_SC->CLKSRCSEL = 1;

	/* Set up PLL0 multiplier and pre-divisor */
	LPC_SC->PLL0CFG  = ((PLL_PREDIV-1) << 16) | (PLL_MULTIPLIER-1);
	LPC_SC->PLL0FEED = 0xaa;
	LPC_SC->PLL0FEED = 0x55;

	/* Enable PLL0 */
	LPC_SC->PLL0CON  = 1;
	LPC_SC->PLL0FEED = 0xaa;
	LPC_SC->PLL0FEED = 0x55;

	/* Increase CPU clock divider */
	LPC_SC->CCLKCFG = PLL_DIVISOR-1;

	/* Wait until PLL locks */
	while (!(LPC_SC->PLL0STAT & BV(26))) ;

	/* Connect PLL0 */
	LPC_SC->PLL0CON  = 3;
	LPC_SC->PLL0FEED = 0xaa;
	LPC_SC->PLL0FEED = 0x55;
}

/* Put MCU in low-power mode */
void system_sleep(void)
{
	__WFI();
}

/* Reset MCU */
void system_reset(void)
{
	__disable_irq();

	/* force watchdog reset */
	LPC_WDT->WDTC = 256;            // minimal timeout
	LPC_WDT->WDCLKSEL = BV(31);     // internal RC, lock register
	LPC_WDT->WDMOD = BV(0) | BV(1); // enable watchdog and reset-by-watchdog
	LPC_WDT->WDFEED = 0xaa;
	LPC_WDT->WDFEED = 0x55;         // initial feed to really enable WDT

	while (1) ;
}

/* ----- crash detection ----- */

void HardFault_Handler(void)
{
	set_led2(1);
	while (1) ;
}

void MemManage_Handler(void)
{
	set_led2(1);
	while (1) ;
}

void BusFault_Handler(void)
{
	set_led2(1);
	while (1);
}

void UsageFault_Handler(void)
{
	set_led2(1);
	while (1) ;
}
