/*
 * ARM JTAG Fault Injector
 *
 * Author: Ingo Korb <ingo@akana.de>
 *
 * Copyright (C) 2011-2014 Department of Computer Science,
 * Design Automation of Embedded Systems Group,
 * Dortmund University of Technology, all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither the name of the copyright holder nor the names of its
 *      contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "hwconfig.h"
#include <stdio.h>
#include <string.h>
#include "arm_opcodes.h"
#include "embice.h"
#include "jtag.h"
#include "timer.h"
#include "target.h"

/* --- macros --- */

#define CPSR_T (1<<5)

/* --- internal state variables --- */

static arm_regs_t halt_regs;
static uint16_t   current_regmask;
static uint32_t   dirty_regs;  // bit 16: CPSR

/* --- helper functions --- */

// table is in jtag.c

/* flip bit order in a word of up to 32 bit length */
static inline uint32_t flip_u32(uint32_t value, unsigned int num) {
  uint32_t out;

  asm("rbit %[result], %[input]" : [result] "=r" (out) : [input] "r" (value));
  return out >> (32 - num);
}

// same as above, but generates smaller code when the argument is a compile-time constant
// FIXME: Unused, remove?
#define flip_u32_static(val, bits) ({                           \
  unsigned int v = (val);                                       \
    v = ((v >> 1) & 0x55555555) | ((v & 0x55555555) << 1);      \
    v = ((v >> 2) & 0x33333333) | ((v & 0x33333333) << 2);      \
    v = ((v >> 4) & 0x0F0F0F0F) | ((v & 0x0F0F0F0F) << 4);      \
    v = ((v >> 8) & 0x00FF00FF) | ((v & 0x00FF00FF) << 8);      \
    v = ( v >> 16             ) | ( v               << 16);     \
    v = v >> (32 - (bits));                                     \
    v;                                                          \
    })

/* feed an instruction with data to the ARM core, special Feroceon version */
static uint32_t core_feed_insn_noidle(uint32_t insn, uint32_t data, unsigned int sysspd) {
  uint32_t i, dout = 0;
  uint8_t s;
  jtag_block_t bl[3] = {
    { 32, &data, &dout },
    {  3, &s,    NULL  },
    { 32, &i,    NULL  }
  };

  /* switch to CPU core scan chain */
  jtag_select_chain(CHAIN_CORE, JTST_PAUSE_DR);
  jtag_set_ir(JTAG_INTEST, JTAG_IR_LEN, JTST_PAUSE_DR);

  /* flip instruction bit order */
  // FIXME: Maybe move to caller or even the macros?
  i = flip_u32(insn, 32);
  if (sysspd)
    s = 4;
  else
    s = 0;

  /* execute */
  jtag_scan_dr(3, bl, JTST_PAUSE_DR);

  return dout;
}

/* feed an instruction with data to the ARM core, standard version */
static uint32_t core_feed_insn(uint32_t insn, uint32_t data, unsigned int sysspd) {
  uint32_t dout = core_feed_insn_noidle(insn, data, sysspd);

  /* clock core */
  jtag_change_state(JTST_PAUSE_DR, 1);

  return dout;
}

/* read data from the ARM core */
static uint32_t core_read(void) {
  uint32_t data;
  jtag_block_t bl[3] = {
    { 32, NULL, &data },
    {  3, NULL, NULL },
    { 32, NULL, NULL }
  };

  /* switch to CPU core scan chain */
  jtag_select_chain(CHAIN_CORE, JTST_PAUSE_DR);
  jtag_set_ir(JTAG_INTEST, JTAG_IR_LEN, JTST_PAUSE_DR);

  /* execute */
  jtag_scan_dr(3, bl, JTST_IDLE);

  return data;
}

/* execute a system-speed instruction */
/* returns 0 if successful, nonzero if not */
static int core_execute_sysspeed(void) {
  // no bypass!
  jtag_set_ir(JTAG_RESTART, JTAG_IR_LEN, JTST_IDLE);

  /* wait until access is completed */
  uint32_t status;
  tick_t timeout = getticks() + MS_TO_TICKS(1000);
  do {
    status = embice_read_register(EMBICE_DEBUG_STATUS);
  } while ((status & (EMBICE_DBGACK | EMBICE_SYSCOMP)) != (EMBICE_DBGACK | EMBICE_SYSCOMP) &&
           !time_after(getticks(), timeout));

  if (time_after(getticks(), timeout)) {
    /* timeout */
    printf("ERROR: execute_sysspeed timeout, status %08lx\n",status);
    return 1;
  }

  return 0;
}

/* read the registers of the target CPU, current mode only */
static void internal_read_registers(arm_regs_t *regs, unsigned int mask, unsigned int read_cpsr) {
  unsigned int i;

  /* current registers */
  core_feed_insn(ARMV4_5_STMIA(0, mask & 0xffff, 0, 0), 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
  for (i=0; i<=15; i++) {
    if (mask & (1 << i))
      regs->r[i] = core_read();
  }
  core_feed_insn(ARMV4_5_NOP, 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);

  if (read_cpsr) {
    /* CPSR */
    // FIXME: Extract - would also work for SPSR with a minor change
    core_feed_insn(ARMV4_5_MRS(0, 0), 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);

    core_feed_insn(ARMV4_5_STMIA(0, 1, 0, 0), 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    regs->cpsr = core_feed_insn(ARMV4_5_NOP, 0, 0);

    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
  }
  if (read_cpsr) {
    /* SPSR */
    // FIXME: Extract - would also work for SPSR with a minor change
    core_feed_insn(ARMV4_5_MRS(0, 1), 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);

    core_feed_insn(ARMV4_5_STMIA(0, 1, 0, 0), 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    regs->spsr = core_feed_insn(ARMV4_5_NOP, 0, 0);

    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
  }
}

/* (partially) updates CPSR (reg=0) or SPSR (reg=1) - mask selects the byte(s) to update */
static void internal_write_xpsr(unsigned int reg, uint32_t value, uint8_t mask) {
  unsigned int i;

  /* lowest byte */
  if (mask & 1) {
    core_feed_insn(ARMV4_5_MSR_IM(value & 0xff, 0, 1, reg), 0, 0);
    for (i=0; i<6; i++)
      core_feed_insn(ARMV4_5_NOP, 0, 0);
  }

  if (mask & 2) {
    core_feed_insn(ARMV4_5_MSR_IM((value & 0xff00) >> 8, 0xc, 2, reg), 0, 0);
    for (i=0; i<6; i++)
      core_feed_insn(ARMV4_5_NOP, 0, 0);
  }

  if (mask & 4) {
    core_feed_insn(ARMV4_5_MSR_IM((value & 0xff0000) >> 16, 0x8, 4, reg), 0, 0);
    for (i=0; i<6; i++)
      core_feed_insn(ARMV4_5_NOP, 0, 0);
  }

  /* highest byte */
  if (mask & 8) {
    core_feed_insn(ARMV4_5_MSR_IM((value & 0xff000000) >> 24, 0x4, 8, reg), 0, 0);
    for (i=0; i<6; i++)
      core_feed_insn(ARMV4_5_NOP, 0, 0);
  }
}

/* set the contents of the target CPU registers (current mode only) */
static void internal_write_registers(arm_regs_t *regs, unsigned int mask, unsigned int write_cpsr) {
  unsigned int i;

  if (write_cpsr) {
    /* remove Thumb mode bit */
    internal_write_xpsr(0, regs->cpsr & ~CPSR_T, 0xf);
  }

  core_feed_insn(ARMV4_5_LDMIA(0, mask & 0xffff, 0, 0), 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);

  for (i=0; i<=15; i++) {
    if (mask & (1 << i))
      core_feed_insn(ARMV4_5_NOP, regs->r[i], 0);
  }

  core_feed_insn(ARMV4_5_NOP, 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
}

static void write_pc(uint32_t pc) {
  core_feed_insn(ARMV4_5_LDMIA(0, 0x8000, 0, 0), 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
  core_feed_insn(ARMV4_5_NOP, pc, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 0);
}

/* --- exported functions --- */

uint32_t target_read_memory(uint32_t address) {
  arm_regs_t tmpregs;

  /* set r0 to target address */
  tmpregs.r[0] = address;
  internal_write_registers(&tmpregs, 0x0001, 0);

  /* tell core to do a system-speed LDM */
  core_feed_insn(ARMV4_5_LDMIA(0, 0x0001, 0, 0), 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 1);

  /* execute */
  // FIXME: Handle timeouts
  core_execute_sysspeed();

  /* read result from register */
  internal_read_registers(&tmpregs, 0x0001, 0);

  return tmpregs.r[0];
}

// FIXME: Allow smaller writes?
void target_write_memory(uint32_t address, uint32_t value) {
  arm_regs_t tmpregs;

  if (!(current_regmask & 2)) {
    printf("ERROR: r1 wasn't saved in target_halt, cannot write to memory!\n");
    return;
  }

  /* set r0 to target address, r1 to value */
  dirty_regs |= 1 << 1;
  tmpregs.r[0] = address;
  tmpregs.r[1] = value;
  internal_write_registers(&tmpregs, 0x0003, 0);

  /* tell core to do a system-speed STM */
  core_feed_insn(ARMV4_5_STMIA(0, 0x0002, 0, 0), 0, 0);
  core_feed_insn(ARMV4_5_NOP, 0, 1);

  /* execute */
  // FIXME: handle timeouts
  core_execute_sysspeed();
}


/* attempts to halt CPU - returns 0 if successful, nonzero if timed out */
int target_halt(uint16_t regmask) {
  unsigned int thumb_mode = 0;
  uint32_t thumb_r0 = 0, thumb_pc = 0;

  /* always save r0 and r15 because they're modified during halt/resume */
  regmask |= 0x8001;
  current_regmask = regmask;
  dirty_regs = 0x8001;

  embice_clear_breaks();
  embice_write_register(EMBICE_DEBUG_CTRL, EMBICE_DBGRQ); // also disables monitor mode

  /* wait until DBGACK is set */
  uint32_t status;
  tick_t timeout = getticks() + MS_TO_TICKS(1000);
  do {
    status = embice_read_register(EMBICE_DEBUG_STATUS);
  } while (!(status & EMBICE_DBGACK) && !time_after(getticks(), timeout));

  if (!(status & EMBICE_DBGACK)) {
    /* timeout */
    return 1;
  }

  embice_write_register(EMBICE_DEBUG_CTRL, EMBICE_DBGACK | EMBICE_INTDIS);

  /* check current CPU mode */
  if (status & EMBICE_IJBIT) {
    printf("***ERROR*** CPU is in Jazelle mode!\n");
    return -1;
  } else if (status & EMBICE_ITBIT) {
    /* switch CPU to ARM */
    thumb_mode = 1;

    core_feed_insn_noidle(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn_noidle(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn_noidle(ARMV4_5_T_NOP, 0, 0);

    core_feed_insn(ARMV4_5_T_STR(0, 0), 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    thumb_r0 = core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);

    core_feed_insn(ARMV4_5_T_MOV(0, 15), 0, 0);
    core_feed_insn(ARMV4_5_T_STR(0, 0), 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    thumb_pc = core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);

    core_feed_insn(ARMV4_5_T_BX(15), 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);

    /* fix Thumb-mode PC */
    thumb_pc -= (12 + 4);
  }

  /* read some registers of the current mode so we can clobber them */
  internal_read_registers(&halt_regs, regmask, 1);

  /* fix PC */
  if (thumb_mode) {
    halt_regs.cpsr |= 1 << 5; // fix T bit
    halt_regs.r[0]  = thumb_r0;
    halt_regs.r[15] = thumb_pc - 3*2;
  } else {
    halt_regs.r[15] -= 3*4 + 3*4; // adjust for STM and DBGRQ
  }

  return 0;
}

/* resumes CPU */
void target_resume(void) {
  uint16_t regmask = dirty_regs & 0x7fff;

  internal_write_registers(&halt_regs, regmask, dirty_regs & (1 << 16));
  write_pc(halt_regs.r[15]);

  if (halt_regs.cpsr & CPSR_T) {
    /* resume to thumb mode */
    uint32_t r0 = halt_regs.r[0];
    uint32_t pc = halt_regs.r[15];

    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);

    core_feed_insn(0xE28F0001, 0, 0); // add r0,pc,#1 - switch to thumb
    core_feed_insn(ARMV4_5_BX(0), 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);

    core_feed_insn(ARMV4_5_T_LDMIA(0, 0x1), 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);

    core_feed_insn(ARMV4_5_T_NOP, r0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);
    core_feed_insn(ARMV4_5_T_NOP, 0, 0);

    pc = (pc & 2) >> 1;
    core_feed_insn(ARMV4_5_T_B(0x7e9 /*+ pc*/), 0, 0);  // FIXME: This works, but I don't know why yet
    core_feed_insn(ARMV4_5_T_NOP, 0, 1);
  } else {
    /* feed the core a NOP/NOP/NOP/B -28/NOP sequence to prepare for restart */
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 0);
    core_feed_insn(ARMV4_5_B(0xfffff9, 0), 0, 0);
    core_feed_insn(ARMV4_5_NOP, 0, 1);
  }

  /* reset INTDIS/DBGACK bits */
  embice_write_register(EMBICE_DEBUG_CTRL, 0); // FIXME: This disables all breakpoints

  /* send BYPASS+RESTART commands */
  jtag_set_ir(JTAG_BYPASS, JTAG_IR_LEN, JTST_IDLE);
  jtag_set_ir(JTAG_RESTART, JTAG_IR_LEN, JTST_IDLE);

  /* wait until restart is confirmed */
  uint32_t status;
  do {
    status = embice_read_register(EMBICE_DEBUG_STATUS);
  } while (status & EMBICE_DBGACK);
}

void target_read_registers(arm_regs_t *regs) {
  /* just copy the registers from halt-state */
  // FIXME: handle state changes
  memcpy(regs, &halt_regs, sizeof(arm_regs_t));
}

void target_write_registers(arm_regs_t *regs) {
  // FIXME: handle state changes
  unsigned int i;

  /* check which values were actually changed */
  for (i=0; i<16; i++) {
    if (halt_regs.r[i] != regs->r[i])
      dirty_regs |= 1 << i;
    halt_regs.r[i] = regs->r[i];
  }

  if (halt_regs.cpsr != regs->cpsr)
    dirty_regs |= 1 << 16;

  halt_regs.cpsr = regs->cpsr;
}

void target_write_register_now(arm_regs_t *regs, uint16_t mask) {
  unsigned int i;

  /* copy new values to halt_regs so they aren't overwritten on resume */
  for (i=0; i<16; i++)
    if ((mask & (1 << i)) || !(current_regmask & (1 << i)))
      halt_regs.r[i] = regs->r[i];

  /* mark updated registers as not dirty */
  dirty_regs &= (1 << 16) | ~mask;

  /* write */
  internal_write_registers(regs, mask, 0);
}

void target_fiq(void) {
  if ((halt_regs.cpsr & 0x1f) == 0x11) {
    // FIXME: Check FIQ-disable-bit too?
    printf("ERROR: target_fiq called while already in FIQ mode!\n");
    return;
  }

  /* restore banked registers if they were changed */
  if (dirty_regs & 0b0111111100000000) {
    target_write_register_now(&halt_regs, dirty_regs & 0b0111111100000000);
    current_regmask  &= 0b1000000011111111;
    dirty_regs       &= 0b1000000011111111;
  }

  /* switch CPU to FIQ mode (write lowest byte of CPSR only) */
  internal_write_xpsr(0, 0b11010001, 1);

  /* set SPSR_fiq */
  internal_write_xpsr(1, halt_regs.cpsr, 0xf);

  /* update halt_regs */
  halt_regs.r[14] = halt_regs.r[15] + 4;
  //halt_regs.r[15] = 0x1c; // FIXME: Ignores exception base
  halt_regs.r[15] = 0xFFFF001c; // FIXME: Ignores exception base
  halt_regs.cpsr  = (halt_regs.cpsr & 0xffffff00) | 0b11010001;

  /* R14 and CPSR need to be "restored" on resume */
  current_regmask |= (1 << 14);
  dirty_regs      |= (1 << 14) | (1 << 16);
}
