/*
 * ARM JTAG Fault Injector
 *
 * Author: Ingo Korb <ingo@akana.de>
 *
 * Copyright (C) 2011-2014 Department of Computer Science,
 * Design Automation of Embedded Systems Group,
 * Dortmund University of Technology, all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither the name of the copyright holder nor the names of its
 *      contributors may be used to endorse or promote products derived from
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "hwconfig.h"
#include <arm/NXP/LPC17xx/LPC17xx.h>
#include <arm/bits.h>
#include <stdio.h>
#include "timer.h"
#include "touchscreen.h"

/* A few symbols to make this code work for all four UARTs */
#if defined(TOUCHSCREEN_UART) && TOUCHSCREEN_UART == 0
#  define UART_PCONBIT 3
#  define UART_PCLKREG PCLKSEL0
#  define UART_PCLKBIT 6
#  define UART_REGS    LPC_UART0
#  define UART_HANDLER UART0_IRQHandler
#  define UART_IRQ     UART0_IRQn
#elif TOUCHSCREEN_UART == 1
#  define UART_PCONBIT 4
#  define UART_PCLKREG PCLKSEL0
#  define UART_PCLKBIT 8
#  define UART_REGS    LPC_UART1
#  define UART_HANDLER UART1_IRQHandler
#  define UART_IRQ     UART1_IRQn
#elif TOUCHSCREEN_UART == 2
#  define UART_PCONBIT 24
#  define UART_PCLKREG PCLKSEL1
#  define UART_PCLKBIT 16
#  define UART_REGS    LPC_UART2
#  define UART_HANDLER UART2_IRQHandler
#  define UART_IRQ     UART2_IRQn
#elif TOUCHSCREEN_UART == 3
#  define UART_PCONBIT 25
#  define UART_PCLKREG PCLKSEL1
#  define UART_PCLKBIT 18
#  define UART_REGS    LPC_UART3
#  define UART_HANDLER UART3_IRQHandler
#  define UART_IRQ     UART3_IRQn
#else
#  error TOUCHSCREEN_UART is not set or has an invalid value!
#endif

static unsigned int baud2divisor(unsigned int baudrate) {
	return CONFIG_MCU_FREQ / 16 / baudrate;
}

static enum {
	STATE_WAIT4START,
	STATE_GOTSTART,
	STATE_BYTE1,
	STATE_BYTE2,
	STATE_BYTE3,
} parser_state = STATE_WAIT4START;

static volatile uint32_t coordinate;
volatile unsigned int coord_updated;
static uint8_t buffer[5];

/* parse received touchscreen data */

static void parse_ts_packet(uint8_t *buffer)
{
	/* status byte */
	// FIXME: Do something useful with this?
	// 0: 0=lift-off point, 1=touch point
	// 2-1: precision
	//       00 - 11 bit
	//       01 - 12 bit
	//       10 - 13 bit
	//       11 - 14 bit
	// 5: byte 5 is player ID
	// 6: byte 5 is pressure
	set_led0(buffer[0] & 1);
	set_led1(buffer[0] & 2);
	set_led2(buffer[0] & 4);
	set_led3(buffer[0] & 0b01100000);

	/* coordinates */
	uint32_t tmp;
	tmp = (buffer[1] << (7+16)) | (buffer[2] << 16) |
			(buffer[3] << 7) | buffer[4];

	if (buffer[0] & 1)
		tmp |= (1UL << 31);

	coordinate = tmp;
	coord_updated = 1;
}

static void touchscreen_parse(uint8_t byte)
{
	switch (parser_state)
	{
	case STATE_WAIT4START:
		if (byte & 0x80)
		{
			parser_state = STATE_GOTSTART;
			buffer[0] = byte;
		}
		break;

	case STATE_GOTSTART:
		if (byte & 0x80)
		{
			/* another start byte, stay in this state */
			buffer[0] = byte;
		} else {
			buffer[1] = byte;
			parser_state = STATE_BYTE1;
		}
		break;

	case STATE_BYTE1:
		if (byte & 0x80)
		{
			/* another start byte, reset */
			buffer[0] = byte;
			parser_state = STATE_GOTSTART;
		} else {
			buffer[2] = byte;
			parser_state = STATE_BYTE2;
		}
		break;

	case STATE_BYTE2:
		if (byte & 0x80)
		{
			/* another start byte, reset */
			buffer[0] = byte;
			parser_state = STATE_GOTSTART;
		} else {
			buffer[3] = byte;
			parser_state = STATE_BYTE3;
		}
		break;

	case STATE_BYTE3:
		if (byte & 0x80)
		{
			/* another start byte, reset */
			buffer[0] = byte;
			parser_state = STATE_GOTSTART;
		}
		else
		{
			buffer[4] = byte;
			parse_ts_packet(buffer);
			parser_state = STATE_WAIT4START;
		}
		break;

	default:
		/* Bug */
		parser_state = STATE_WAIT4START;
		break;
	}
}

void UART_HANDLER(void)
{
	int iir = UART_REGS->IIR;

	if (!(iir & 1))
	{
		/* Interrupt is pending */
		switch (iir & 14)
		{
#if UART_NUMBER == 1
		case 0: /* modem status */
			(void) UART_REGS->MSR; // dummy read to clear
			break;
#endif

		case 2: /* THR empty - turn off THRE int, this code is rx-only */
			BITBAND(UART_REGS->IER, 1) = 0;
			break;

		case  4: /* data received - not implemented yet */
			touchscreen_parse(UART_REGS->RBR);
			break;

		case 12: /* RX timeout */
			(void) UART_REGS->RBR; // dummy read to clear
			break;

		case 6: /* RX error */
			(void) UART_REGS->LSR; // dummy read to clear
			break;

		default:
			break;
		}
	}
}

void touchscreen_toggle_int(void)
{
	TOUCHSCREEN_INT_GPIO->FIOCLR = BV(TOUCHSCREEN_INT_PIN);
	delay_ms(1);
	TOUCHSCREEN_INT_GPIO->FIOSET = BV(TOUCHSCREEN_INT_PIN);
}

uint32_t touchscreen_get(void)
{
	coord_updated = 0;
	return coordinate;
}

void touchscreen_init(void)
{
	/* initialize TSINT-GPIO */
	TOUCHSCREEN_INT_GPIO->FIOSET = BV(TOUCHSCREEN_INT_PIN);
	TOUCHSCREEN_INT_GPIO->FIODIR |= BV(TOUCHSCREEN_INT_PIN);

	/* Turn on power to UART */
	BITBAND(LPC_SC->PCONP, UART_PCONBIT) = 1;

	/* UART clock = CPU clock */
	BITBAND(LPC_SC->UART_PCLKREG, UART_PCLKBIT  ) = 1;
	BITBAND(LPC_SC->UART_PCLKREG, UART_PCLKBIT+1) = 0;

	/* set baud rate - no fractional stuff for now */
	UART_REGS->LCR = BV(7) | 3; // always 8n1
	UART_REGS->DLL = baud2divisor(TOUCHSCREEN_BAUD) & 0xff;
	UART_REGS->DLM = (baud2divisor(TOUCHSCREEN_BAUD) >> 8) & 0xff;
	BITBAND(UART_REGS->LCR, 7) = 0;

	/* reset and enable FIFO */
	UART_REGS->FCR = BV(0);

	/* connect RX pin */
	tsuart_connect();

	/* enable receive interrupt */
	BITBAND(UART_REGS->IER, 0) = 1;
	NVIC_EnableIRQ(UART_IRQ);
}
